// (C) 2001-2020 Intel Corporation. All rights reserved. 
// Your use of Intel Corporation's design tools, logic functions and other 
// software and tools, and its AMPP partner logic functions, and any output 
// files from any of the foregoing (including device programming or simulation 
// files), and any associated documentation or information are expressly subject 
// to the terms and conditions of the Intel Program License Subscription 
// Agreement, Intel FPGA IP License Agreement, or other applicable 
// license agreement, including, without limitation, that your use is for the 
// sole purpose of programming logic devices manufactured by Intel and sold by 
// Intel or its authorized distributors.  Please refer to the applicable 
// agreement for further details. 

#include <errno.h>
#include <io.h>
#include <stddef.h>
#include <strings.h>
#include "alt_types.h"
#include "intel_generic_serial_flash_interface_top.h"
#include "intel_generic_serial_flash_interface_top_regs.h"
#include "priv/alt_busy_sleep.h"
#include "sys/alt_cache.h"
#include "sys/alt_debug.h"
#include "sys/param.h"

#ifdef SAFECLIB
#include "safe_mem_lib.h"
#endif

/* 0.7 sec timeout */
#define INTEL_GSFI_1US_TIMEOUT_VALUE                   700000

/*************************************
 * Flash parameters (Device dependent)
 * This is parameter is for EPCQ flash
 * ***********************************
 */
#define CMD_READ_OPERATION                             0x03
#define CMD_WRITE_ENABLE                               0x06
#define CMD_WRITE_OPERATION                            0x01
#define CMD_ERASE_SECTOR                               0xD8
#define CMD_STATUS_REGISTER                            0x05
#define CMD_FLAG_STATUS_REGISTER                       0x70
#define CMD_PAGE_PROGRAM                               0x02
#define CMD_4BYTES_ADDR_MODE                           0xB7

#define FLASH_STATUS_REGISTER_WIP_MSK                  (1)
#define FLAG_STATUS_ILLEGAL_WRITE_MSK                  (1 << 4)
#define FLAG_STATUS_ERASE_FAIL_MSK                     (1 << 5)


/*
 *** Private Function
 *** Helper functions used by Public API functions.***
 *****************************************************
 */

/* intel_gsfi_validate_read_write_arguments
 *
 * Use to check the argument read and write are valid
 *
 *** Arguments:
 * - *flash_info: Pointer to QSPI flash device structure.
 * - offset     : Offset of read/write from base address.
 * - length     : Length of read/write in bytes.
 *
 * Returns:
 * 0 -> success
 * -EINVAL -> Invalid arguments
 */
static alt_32 intel_gsfi_validate_read_write_arguments(intel_gsfi_dev *flash_info, 
                                                       alt_u32 offset, int length)       
{
    alt_u32 start_address = 0;
    alt_u32 end_address   = 0;

    start_address = flash_info->data_base + offset;
    end_address   = start_address + length;

    if(start_address > flash_info->data_end ||
         end_address > flash_info->data_end ||
         offset < 0                         ||
         length <= 0)
    {
        return -EINVAL;
    }

    return 0;
}


/* intel_gsfi_poll_for_write_in_progress
 *
 * Private function that polls write in progress bit polling register.
 *
 * Write in progress will be set if any of the following operations are in progress:
 *     -WRITE STATUS REGISTER
 *     -WRITE NONVOLATILE CONFIGURATION REGISTER
 *     -PROGRAM
 *     -ERASE
 *
 * Assumes QSPI was configured correctly.
 *
 * If INTEL_GSFI_1US_TIMEOUT_VALUE is set, the function will time out after
 * a period of time determined by that value.
 *
 * Arguments:
 * - *qspi_flash_info: Pointer to QSPI flash device structure.
 *
 * Returns:
 *  0 -> success
 * -EINVAL -> Invalid arguments
 * -ETIME  -> Time out and skipping the looping after 0.7 sec.
 *
 */
static alt_32 intel_gsfi_poll_for_write_in_progress(intel_gsfi_dev *flash_info)
{
    alt_u32 cmd;

    /* we'll want to implement timeout if a timeout value is specified */
#if INTEL_GSFI_1US_TIMEOUT_VALUE > 0
    alt_u32 timeout = INTEL_GSFI_1US_TIMEOUT_VALUE;
    alt_u32 counter = 0;
#endif

    if(NULL == flash_info)
    {
        return -EINVAL;
    }

    do{
        cmd = ((1 << INTEL_GSFI_CMD_NUM_DATA_BYTES_OFST)    |
                     INTEL_GSFI_CMD_READ_TYPE_MSK |
                     CMD_STATUS_REGISTER);
        IOWR_INTEL_GSFI_FLASH_CMD(flash_info->csr_base, cmd);
        IOWR_INTEL_GSFI_START_CMD(flash_info->csr_base, INTEL_GSFI_START_CMD_MSK);
      }
    while((IORD_INTEL_GSFI_READ_DATA0(flash_info->csr_base) &
           FLASH_STATUS_REGISTER_WIP_MSK));
        {
            if(counter > (INTEL_GSFI_1US_TIMEOUT_VALUE >> 1))
            {
                alt_busy_sleep(2); /* delay 2us */
            }

#if INTEL_GSFI_1US_TIMEOUT_VALUE > 0
            if(timeout <= counter)
            {
                return -ETIME;
            }
            counter++;
#endif
        }

    return 0;
}


/**
 * intel_gsfi_write_block
 *
 * This function writes one block/sector of data to flash. The length of the 
 * write can NOT spill into the adjacent sector.
 *
 * It assumes that someone has already erased the appropriate sector(s).
 *
 * Arguments:
 * - *flash_info: Pointer to QSPI flash device structure.
 * - block_offset: byte-addressed offset, from the start of flash, of the sector
 *   to written to
 * - data-offset: Byte offset (unaligned access) of write into flash memory.
 *                For best performance, word(32 bits - aligned access) offset of
 *                write is recommended.
 * - *src_addr: source buffer
 * - length: size of writing
 *
 * Returns:
 * 0 -> success
 * -EINVAL -> Invalid arguments
 * -EIO -> write failed, sector might be protected
**/
int intel_gsfi_write_block (alt_flash_dev *flash_info, int block_offset,
                            int data_offset, const void *data, int length)
{
    alt_u32 buffer_offset    = 0; /** offset into data buffer to get write data */
    alt_u32 remaining_length = length; /** length left to write */
    alt_u32 write_offset     = data_offset; /** offset into flash to write too */
    alt_u32 cmd;
    intel_gsfi_dev* qspi_flash_info = NULL;
    int ret_code;

    qspi_flash_info = (intel_gsfi_dev*) flash_info;

    /*
     * Sanity checks that data offset is not larger then a sector, that block offset is
     * sector aligned and within the valid flash memory range and a write doesn't spill into
     * the adjacent flash sector.
     */
    if( block_offset <  0   ||
        data_offset  <  0   ||
        length       <= 0   ||
        NULL == data        ||
        NULL == flash_info  ||
        data_offset  >= qspi_flash_info->size_in_bytes  ||
        block_offset >= qspi_flash_info->size_in_bytes  ||
        ((block_offset & (qspi_flash_info->sector_size - 1)) != 0 ) ||
        length > (qspi_flash_info->sector_size - (data_offset - block_offset)))
    {
        return -EINVAL;
    }


    ret_code = intel_gsfi_poll_for_write_in_progress(qspi_flash_info);

    if(0 != ret_code)
    {
        return -EINVAL;
    }
    
    /* Setup Operating Protocol and Write Instruction Register */    
    cmd = ((CMD_FLAG_STATUS_REGISTER << 8) | CMD_PAGE_PROGRAM);

    IOWR_INTEL_GSFI_PROTOCOL(qspi_flash_info->csr_base, 0x0);
    IOWR_INTEL_GSFI_WRITE_INST(qspi_flash_info->csr_base, cmd);
            
    /*
     * Do writes one 32-bit word at a time.
     * We need to make sure that we pad the first few bytes so they're word aligned if they are
     * not already.
     */

    while(remaining_length > 0)
    {
        alt_u32 word_to_write = 0xFFFFFFFF; /** initialize word to write to blank word */
        alt_u32 padding = 0; /** bytes to pad the next word that is written */
        alt_u32 bytes_to_copy = sizeof(alt_u32);  /** number of bytes from source to copy */

        /*
         * we need to make sure the write is word aligned
         * this should only be true at most 1 time
         */
        if(0 != (write_offset & (sizeof(alt_u32) - 1)))
        {
             /*
             * data is not word aligned
             * calculate padding bytes need to add before start of a data offset
             */
            padding = write_offset & (sizeof(alt_u32) - 1);

            /* update variables to account for padding being added */
            bytes_to_copy -= padding;

            if(bytes_to_copy > remaining_length)
            {
                bytes_to_copy = remaining_length;
            }

            write_offset = write_offset - padding;

            if(0 != (write_offset & (sizeof(alt_u32) - 1)))
            {
                return -EINVAL;
            }
        }
        else
        {
            if(bytes_to_copy > remaining_length)
            {
                bytes_to_copy = remaining_length;
            }
        }

#ifdef SAFECLIB
        ret_code = memcpy_s(((void*)&word_to_write) + padding, bytes_to_copy,
                            ((void*)data) + buffer_offset, bytes_to_copy);

        if(0 != ret_code)
        {
            return ret_code;
        }

#else
#error Missing altera safeclib. Please enable it in BSP Software Package.
#endif
   
        /* update offset and length variable */
        buffer_offset    += bytes_to_copy;
        remaining_length -= bytes_to_copy;
   
        IOWR_32DIRECT(qspi_flash_info->data_base, write_offset, word_to_write);
  
        /*update current offset */
        write_offset += sizeof(alt_u32);

    }

    ret_code = intel_gsfi_poll_for_write_in_progress(qspi_flash_info);

    if(0 != ret_code)
    {
        return -EINVAL;
    }
    
    /* Check flag status register */
    cmd = ((1 << INTEL_GSFI_CMD_NUM_DATA_BYTES_OFST)    |
                 INTEL_GSFI_CMD_READ_TYPE_MSK |
                 CMD_FLAG_STATUS_REGISTER);
    IOWR_INTEL_GSFI_FLASH_CMD(qspi_flash_info->csr_base, cmd);
    IOWR_INTEL_GSFI_START_CMD(qspi_flash_info->csr_base, INTEL_GSFI_START_CMD_MSK);

    if((IORD_INTEL_GSFI_READ_DATA0(qspi_flash_info->csr_base) &
        FLAG_STATUS_ILLEGAL_WRITE_MSK))
    {
        return -EIO;
    }
    
    return 0;
}


/**
  * intel_gsfi_erase_block
  *
  * This function erases a single flash sector.
  *
  * Arguments:
  * - *flash_info: Pointer to QSPI flash device structure.
  * - block_offset: byte-addressed offset, from start of flash, of the sector 
  * to be erased
  *
  * Returns:
  * 0 -> success
  * -EINVAL -> Invalid arguments
  * -EIO -> write failed, sector might be protected
**/
int intel_gsfi_erase_block (alt_flash_dev *flash_info, int block_offset)
{
    intel_gsfi_dev *qspi_flash_info = NULL;
    alt_u32 cmd;
    int ret_code;

    if(NULL == flash_info)
    {
        return -EINVAL;
    }

    qspi_flash_info = (intel_gsfi_dev*) flash_info;

    if(block_offset < 0 ||
      (block_offset > (qspi_flash_info->data_end - qspi_flash_info->data_base)) ||
      ((block_offset & (qspi_flash_info->sector_size - 1)) != 0))
    {
        return -EINVAL;
    }

    ret_code = intel_gsfi_poll_for_write_in_progress(qspi_flash_info);

    if(0 != ret_code)
    {
        return ret_code;
    }

    if(qspi_flash_info->addr_4byte_mode == 1)
    {
         cmd = ((4 << INTEL_GSFI_CMD_NUM_ADDR_BYTES_OFST) | CMD_ERASE_SECTOR);
    }
    else
    {
         cmd = ((3 << INTEL_GSFI_CMD_NUM_ADDR_BYTES_OFST) | CMD_ERASE_SECTOR);
    }

    IOWR_INTEL_GSFI_FLASH_CMD(qspi_flash_info->csr_base, CMD_WRITE_ENABLE);
    IOWR_INTEL_GSFI_START_CMD(qspi_flash_info->csr_base,
                              INTEL_GSFI_START_CMD_MSK);
    IOWR_INTEL_GSFI_FLASH_CMD(qspi_flash_info->csr_base, cmd);
    IOWR_INTEL_GSFI_ADDR_REG(qspi_flash_info->csr_base, block_offset);
    IOWR_INTEL_GSFI_START_CMD(qspi_flash_info->csr_base,
                              INTEL_GSFI_START_CMD_MSK);
 
    ret_code = intel_gsfi_poll_for_write_in_progress(qspi_flash_info);

    if(0 != ret_code)
    {
        return -EINVAL;
    }

    /* Check flag status register */
    cmd = ((1 << INTEL_GSFI_CMD_NUM_DATA_BYTES_OFST)    |
           INTEL_GSFI_CMD_READ_TYPE_MSK |
           CMD_FLAG_STATUS_REGISTER);
    IOWR_INTEL_GSFI_FLASH_CMD(qspi_flash_info->csr_base, cmd);
    IOWR_INTEL_GSFI_START_CMD(qspi_flash_info->csr_base, INTEL_GSFI_START_CMD_MSK);

    if((IORD_INTEL_GSFI_READ_DATA0(qspi_flash_info->csr_base) &
        FLAG_STATUS_ERASE_FAIL_MSK))
    {
        return -EIO;
    }

    return 0;
}


/**
 * intel_gsfi_write
 *
 * Program the data into the flash at the selected address.
 *
 * The different between this function and intel_gsfi_write_block function
 * is that this function (intel_gsfi_write) will automatically erase a block as needed
 * Arguments:
 * - *flash_info: Pointer to QSPI flash device structure.
 * - offset: Byte offset (unaligned access) of write to flash memory. 
 * For best performance, word(32 bits - aligned access) offset of write 
 * is recommended.
 * - *src_addr: source buffer
 * - length: size of writing
 *
 * Returns:
 * 0 -> success
 * -EINVAL -> Invalid arguments
 * -EIO -> write failed, sector might be protected
 *
**/
int intel_gsfi_write(alt_flash_dev *flash_info, int offset, const void *src_addr,
                     int length)
{
    alt_u32 write_offset     = offset;
    alt_u32 remaining_length = length;
    alt_u32 buffer_offset    = 0;
    alt_u32 i                = 0;
    int ret_code;

    intel_gsfi_dev* qspi_flash_info = NULL;

    /* return -EINVAL if flash_info and src_addr are NULL */
    if(NULL == flash_info || NULL == src_addr)
    {
        return -EINVAL;
    }

    qspi_flash_info = (intel_gsfi_dev*) flash_info;

    /* make sure the write parameters are within the bounds of the flash */
    ret_code = intel_gsfi_validate_read_write_arguments(qspi_flash_info, offset, length);

    if(0 != ret_code)
    {
        return -EINVAL;
    }

    /*
    * This loop erases and writes data one sector at a time. We check for write completion
    * before starting the next sector.
    */
    for(i = offset/qspi_flash_info->sector_size; i < qspi_flash_info->number_of_sectors; i++)
    {
        alt_u32 block_offset = 0; /** block offset in byte addressing */
        alt_u32 offset_within_current_sector = 0; /** offset into current sector to write */
        alt_u32 length_to_write = 0; /** length to write to current sector */

        if(0 >= remaining_length)
        {
            break; /* out of data to write */
        }

        /* calculate current block/sector offset in byte addressing */
        block_offset = write_offset & (~(qspi_flash_info->sector_size - 1));

        if(block_offset != write_offset)
        {
            offset_within_current_sector = write_offset - block_offset;
        }

        /* erase sector */
        ret_code = intel_gsfi_erase_block(flash_info, block_offset);

        if(0 != ret_code)
        {
            return ret_code;
        }

        /* calculate the byte size of data to be written in a sector */
        length_to_write = MIN(qspi_flash_info->sector_size - offset_within_current_sector,
                              remaining_length);

        /* write data to erased block */
        ret_code = intel_gsfi_write_block(flash_info, block_offset, write_offset,
                                          src_addr + buffer_offset, length_to_write);

        if(0 != ret_code)
        {
            return ret_code;
        }

        /* update remaining length and buffer offset pointer */
        remaining_length -= length_to_write;
        buffer_offset    += length_to_write;
        write_offset     += length_to_write;
    }

    return ret_code;
}


/*
 * intel_gsfi_init
 *
 * alt_sys_init.c will call this function automatically through macro
 *
 * if the flash density input by user is larger than 128Mb, 4-byte addressing
 * mode will be initialize.
 *
 * Arguments:
 * - *dev : Pointer to the QSPI flash device structure.
 *
 *  Returns:
 *  - 0 -> success
 *  -EINVAL -> Invalid argument
 *  -ENODEV -> System is configure incorrectly
 */

alt_u32 intel_gsfi_init(intel_gsfi_dev *flash)
{
    alt_u32 cmd;

    /* return -EINVAL if dev is NULL */
    if(NULL == flash)
    {
        return -EINVAL;
    }

     /* return -ENODEV if CSR slave is not attached */
    if(NULL == (void *)flash->csr_base)
    {
        return -ENODEV;
    }

    /*
     * populate fields of region_info required to conform to HAL API
     * create 1 region that composed of "number_of_sectors" blocks
     */
    flash->dev.number_of_regions     = 1;
    flash->dev.region_info[0].offset = 0;
    flash->dev.region_info[0].region_size      = flash->size_in_bytes;
    flash->dev.region_info[0].number_of_blocks = flash->number_of_sectors;
    flash->dev.region_info[0].block_size = flash->sector_size;

    /*
     * Assumption:
     **** Flash density insert correctly ***
     *
     * Register this device as a valid flash device
     */

    if(flash->size_in_bytes > 0xFFFFFF)
    {
        IOWR_INTEL_GSFI_FLASH_CMD(flash->csr_base, CMD_WRITE_ENABLE);
        IOWR_INTEL_GSFI_START_CMD(flash->csr_base, INTEL_GSFI_START_CMD_MSK);
        IOWR_INTEL_GSFI_FLASH_CMD(flash->csr_base, CMD_4BYTES_ADDR_MODE);
        IOWR_INTEL_GSFI_START_CMD(flash->csr_base, INTEL_GSFI_START_CMD_MSK);

        cmd = (INTEL_GSFI_CTR_ADDR_MODE_MSK |
               INTEL_GSFI_CTR_EN_MSK);

        IOWR_INTEL_GSFI_CTR(flash->csr_base, cmd);
        flash->addr_4byte_mode = 1;
    }

    alt_flash_device_register(&(flash->dev));

    return 0;
}


/**
 * intel_gsfi_get_info
 *
 * Pass the table of erase blocks to the user. This flash will return a single
 * flash_region that gives the number and size of sectors for the device used.
 *
 * Arguments:
 * - *fd: Pointer to general flash device structure.
 * - **info: Pointer to flash region
 * - *number_of_regions: Pointer to number of regions
 *
 * For details of setting sectors protection, please refer to EPCQ/QSPI datasheet.
 *
 * Returns:
 * 0 -> success
 * -EINVAL -> Invalid arguments
 * -EIO -> Could be hardware problem.
**/

int intel_gsfi_get_info(alt_flash_dev *fd, flash_region **info, 
                        int *number_of_regions)
{
    alt_flash_dev* flash = NULL;

    if(NULL == fd || NULL == info || NULL == number_of_regions)
    {
        return -EINVAL;
    }

    flash = (alt_flash_dev*)fd;

    *number_of_regions = flash->number_of_regions;

    if(!flash->number_of_regions)
    {
        return -EIO;
    }
    else
    {
        *info = &flash->region_info[0];
    }

    return 0;
}


/*
 * intel_gsfi_read
 *
 * There's no real need to use this function as opposed to using memcpy directly.
 * It does do some sanity checks on the bounds of the read.
 *
 * Arguments:
 * - *flash_info: Pointer to general flash device structure.
 * - offset: offset read from flash memory.
 * - *dest_addr: destination buffer
 * - length: size of reading
 *
 * Returns:
 * 0 -> success
 * -EINVAL -> Invalid arguments
 *
 */

int intel_gsfi_read (alt_flash_dev *flash_info, int offset, void *dest_addr,
                     int length)
{
    int ret_code;
    intel_gsfi_dev *qspi_flash_info = NULL;

    if((NULL == flash_info) || (NULL == dest_addr))
    {
        return -EINVAL;
    }

    qspi_flash_info = (intel_gsfi_dev*)flash_info;
    
    /* Validate Argument */
    ret_code = intel_gsfi_validate_read_write_arguments(qspi_flash_info, offset, length);

    if (0 != ret_code)
    {
        return ret_code;
    }
    
    alt_dcache_flush_no_writeback((alt_u8*)qspi_flash_info->data_base + offset, length);
    
    IOWR_INTEL_GSFI_PROTOCOL(qspi_flash_info->csr_base, 0x0);
    IOWR_INTEL_GSFI_READ_INST(qspi_flash_info->csr_base, CMD_READ_OPERATION);

#ifdef SAFECLIB

        ret_code = memcpy_s(dest_addr, length, 
                           (alt_u8*)qspi_flash_info->data_base + offset, length);

    
#else
#error Missing altera safeclib. Please enable it in BSP Software Package.
#endif
 
    return ret_code;
}
