// (C) 2001-2021 Intel Corporation. All rights reserved.
// Your use of Intel Corporation's design tools, logic functions and other 
// software and tools, and its AMPP partner logic functions, and any output 
// files from any of the foregoing (including device programming or simulation 
// files), and any associated documentation or information are expressly subject 
// to the terms and conditions of the Intel Program License Subscription 
// Agreement, Intel FPGA IP License Agreement, or other applicable 
// license agreement, including, without limitation, that your use is for the 
// sole purpose of programming logic devices manufactured by Intel and sold by 
// Intel or its authorized distributors.  Please refer to the applicable 
// agreement for further details.


// PRS Soft Logic to set prs_res_failure_q, prs_uqrgi_q and prs_stopped_q as expected.
// HSD# 1408732831 

module prs_soft_logic
(
	input clk,
	input rst_n,
	input prs_event_valid_i,
	input [2:0] prs_event_func_i,
	input [1:0] prs_event_i,
	output reg prs_event_valid_o,
	output reg [2:0] prs_event_func_o,
	output reg [1:0] prs_event_o
);

//Parameters
	parameter [1:0] IDLE = 2'h0, PRS_EVENT = 2'h1, PRS_WAIT = 2'h2, PRS_CLEAR = 2'h3;
	parameter DATA_WIDTH = 5, ADDR_WIDTH = 4;
	parameter RAM_DEPTH = (1 << ADDR_WIDTH);

	reg	[1:0]state, prs_counter;
	reg wait_done, wr_en, rd_en ;
	reg [ADDR_WIDTH-1:0] wr_ptr, rd_ptr, status_cnt;
	reg prs_event_valid_w;
	reg [2:0] prs_event_func_w;
	reg [1:0] prs_event_w;

//RAM 
	reg [DATA_WIDTH-1:0] ram[2**ADDR_WIDTH-1:0];
	reg [DATA_WIDTH-1:0] ram_o;

	wire empty, full;
	wire [2:0] func_ram_o;
	wire [1:0] event_ram_o;


	assign func_ram_o = ram_o[4:2];
	assign event_ram_o = ram_o[1:0];

	assign empty = (status_cnt == 4'd0);
	assign full = (status_cnt == (RAM_DEPTH -1));

	always @(posedge clk) begin
		if (!rst_n) begin
			prs_event_valid_o <= '0;
			prs_event_func_o <= '0;
			prs_event_o <= '0;
		end
		else begin
			prs_event_valid_o <= prs_event_valid_w;
			prs_event_func_o <= prs_event_func_w;
			prs_event_o <= prs_event_w;
		end
	end
	
// Output depends only on the state
	always_comb begin : proc_prs_fsm_comb
		case (state)
			IDLE: begin
				prs_event_valid_w = 1'b0;
				prs_event_func_w	= 3'b000;
				prs_event_w 		= 2'b00;
				end
			PRS_EVENT: begin
				prs_event_valid_w	= 1'b1;
				prs_event_func_w	= func_ram_o;
				prs_event_w		= event_ram_o;
				end
			PRS_WAIT: begin
				prs_event_valid_w	= 1'b0;
				prs_event_func_w	= 3'b000;
				prs_event_w		= 2'b00;
				end
			PRS_CLEAR: begin
				prs_event_valid_w 	= 1'b1;
				prs_event_func_w	= func_ram_o;
				prs_event_w		= 2'b11;
				end
			default: begin
				prs_event_valid_w	= 1'b0;
				prs_event_func_w	= 3'b000;
				prs_event_w		= 2'b00;
				end
		endcase
	end

// Determine the next state
	always_ff @ (posedge clk) begin : proc_prs_fsm
		if (!rst_n)
			state <= IDLE;
		else
			case (state)
				IDLE:
					if (!empty)
					state <= PRS_EVENT;
				PRS_EVENT:
					state <= PRS_WAIT;
					
				PRS_WAIT:
					if (wait_done)
						state <= PRS_CLEAR;
					else 
						state <= PRS_WAIT;
				PRS_CLEAR:
					state <= IDLE;
					
			endcase
	end

	always @ (posedge clk ) begin
		if (state == PRS_WAIT)	
			prs_counter <= prs_counter + 1'b1;
		else
			prs_counter <= 2'd0;
	end
	
	always @ (posedge clk) begin
		if (prs_counter == 2'd3) 
			wait_done <= 1'b1;
		else
			wait_done <= 1'b0;
	end
	
	always @ (posedge clk) begin
		if (!rst_n) begin
			wr_ptr	<= 4'd0;
			wr_en	<= 1'b0;
		end else if (prs_event_valid_i) begin
			wr_ptr	<= wr_ptr + 1'b1;
			wr_en	<= 1'b1;
		end else begin 
			wr_ptr	<= wr_ptr;
			wr_en	<= 1'b0;
		end
	end
	
	always @ (posedge clk) begin
		if (!rst_n) begin
			rd_ptr	<= 4'd0;
			rd_en	<= 1'b0;
		end else if (wait_done) begin
			rd_ptr	<= rd_ptr + 1'b1;
			rd_en	<= 1'b1;
		end else begin 
			rd_ptr	<= rd_ptr;
			rd_en	<= 1'b0;
		end
	end
	always @ (posedge clk) begin
		if (!rst_n) begin
			status_cnt	<= 4'd0;
		end else if (wr_en && !rd_en && (status_cnt != RAM_DEPTH)) begin
			status_cnt	<= status_cnt + 1'b1;
		end else if (!wr_en && rd_en && (status_cnt != 0)) begin
			status_cnt	<= status_cnt - 1'b1;
		end else begin 
			status_cnt	<= status_cnt;
		end
	end

//RAM Instance


	always @ (posedge clk)
	begin
		if (prs_event_valid_i)
			ram[wr_ptr] <= {prs_event_func_i,prs_event_i};
		ram_o <= ram[rd_ptr];
	end
	
endmodule
