// (C) 2001-2021 Intel Corporation. All rights reserved.
// Your use of Intel Corporation's design tools, logic functions and other 
// software and tools, and its AMPP partner logic functions, and any output 
// files from any of the foregoing (including device programming or simulation 
// files), and any associated documentation or information are expressly subject 
// to the terms and conditions of the Intel Program License Subscription 
// Agreement, Intel FPGA IP License Agreement, or other applicable 
// license agreement, including, without limitation, that your use is for the 
// sole purpose of programming logic devices manufactured by Intel and sold by 
// Intel or its authorized distributors.  Please refer to the applicable 
// agreement for further details.


// synthesis translate_off
`timescale 1ns / 1ps
// synthesis translate_on

//-----------------------------------------------------------------------------
// Description   : WHR PCI Express VirtIO CFG Read Requests returned 
//                 data flow control and CII outputs State Machine
//-----------------------------------------------------------------------------
// File          : intel_pcie_virtio_dataflow.v
// Author        : Intel Corporation
//-----------------------------------------------------------------------------
//

module intel_pcie_virtio_dataflow #(
    parameter TOTAL_PF_COUNT_WIDTH            = 3,
    parameter TOTAL_VF_COUNT_WIDTH            = 11,
    parameter REG_ADDR_WIDTH                  = 10

)(
        input                             clk,
        input                             rstn,
        // Read Data from VirtIO Common Config Capability
        input  [31:0]                     virtio_cmncfg_header_data_i,
        input  [31:0]                     virtio_cmncfg_bar_data_i,
        input  [31:0]                     virtio_cmncfg_baroffset_data_i,
        input  [31:0]                     virtio_cmncfg_length_data_i,
        // Read Data from VirtIO Notifications Capability
        input  [31:0]                     virtio_notifications_header_data_i,
        input  [31:0]                     virtio_notifications_bar_data_i,
        input  [31:0]                     virtio_notifications_baroffset_data_i,
        input  [31:0]                     virtio_notifications_length_data_i,
        input  [31:0]                     virtio_notifications_notifyoff_data_i,
        // Read Data from VirtIO ISR Status Capability
        input  [31:0]                     virtio_isrstatus_header_data_i,
        input  [31:0]                     virtio_isrstatus_bar_data_i,
        input  [31:0]                     virtio_isrstatus_baroffset_data_i,
        input  [31:0]                     virtio_isrstatus_length_data_i,
        // Read Data from VirtIO Device Specific Capability
        input  [31:0]                     virtio_devspecific_header_data_i,
        input  [31:0]                     virtio_devspecific_bar_data_i,
        input  [31:0]                     virtio_devspecific_baroffset_data_i,
        input  [31:0]                     virtio_devspecific_length_data_i,
        // Read Data from VirtIO PCI Config Access Capability
        input  [31:0]                     virtio_pcicfgaccess_header_data_i,
        input  [31:0]                     virtio_pcicfgaccess_bar_data_i,
        input  [31:0]                     virtio_pcicfgaccess_baroffset_data_i,
        input  [31:0]                     virtio_pcicfgaccess_length_data_i,
        input  [31:0]                     virtio_pcicfgaccess_cfgdata_data_i,
        // CFG Req Ack
        input                             virtio_cfgwr_req_ack_i,
        input                             virtio_cfgrd_req_ack_i,
        // CII HIP Interface
        output                            cii_override_en_o,     // When your application logic asserts this output, the HIP overrides the CfgWr payload 
                                                                 // or CfgRd completion using the data supplied by your application logic on cii_override_din.
        output [31:0]                     cii_override_din_o,
        output                            cii_halt_o,            // When cii_halt_p0/1 is asserted, the HIP halts the processing of CFG reqs for the PCIe config space regs.
        input                             cii_req_i,
        input                             cii_hdr_poisoned_i,
        input  [3:0]                      cii_hdr_first_be_i,
        input  [TOTAL_PF_COUNT_WIDTH-1:0] cii_func_num_i,
        input                             cii_wr_vf_active_i,
        input  [TOTAL_VF_COUNT_WIDTH-1:0] cii_vf_num_i,
        input                             cii_wr_i,
        input  [REG_ADDR_WIDTH-1:0]       cii_addr_i,
        input  [31:0]                     cii_dout_i,
        // CII User Interface
        input                             pin_cii_override_en_i,
        input  [31:0]                     pin_cii_override_din_i,
        input                             pin_cii_halt_i,
        output                            pin_cii_req_o,
        output                            pin_cii_hdr_poisoned_o,
        output [3:0]                      pin_cii_hdr_first_be_o,
        output [TOTAL_PF_COUNT_WIDTH-1:0] pin_cii_func_num_o,
        output                            pin_cii_wr_vf_active_o,
        output [TOTAL_VF_COUNT_WIDTH-1:0] pin_cii_vf_num_o,
        output                            pin_cii_wr_o,
        output [REG_ADDR_WIDTH-1:0]       pin_cii_addr_o,
        output [31:0]                     pin_cii_dout_o,
        // Address Decoder Interface
        input                             targeted_cap_not_present_i,
        input                             noregbank_selected_i   // ~address hit, halt=0, override=0, no change to cfgreq
        
    );
    
    localparam IDLE                 = 2'd0;
    localparam CFGREQ               = 2'd1;
    localparam NOHIT                = 2'd2;
    
    // Signal Declaration
    genvar i;

    reg  [1:0 ]                    cii_ctrl_state;
    reg                            cii_override_en_reg;
    wire                           cii_override_en;
    reg  [31:0]                    cii_override_din_reg;
    wire [31:0]                    cii_override_din;
    reg                            cii_halt_reg;
    wire                           cii_halt;
    reg                            targeted_cap_not_present_reg;  // used for edge detect
    reg                            virtio_cfgwr_req_ack_reg;  // used for edge detect
    reg                            virtio_cfgrd_req_ack_reg;  // used for edge detect
    reg                            noregbank_selected_reg;  // used for edge detect

    reg                            cii_req_reg;
    reg                            cii_hdr_poisoned_reg;
    reg [3:0]                      cii_hdr_first_be_reg;
    reg [TOTAL_PF_COUNT_WIDTH-1:0] cii_func_num_reg;
    reg                            cii_wr_vf_active_reg;
    reg [TOTAL_VF_COUNT_WIDTH-1:0] cii_vf_num_reg;
    reg                            cii_wr_reg;
    reg [REG_ADDR_WIDTH-1:0]       cii_addr_reg;
    reg [31:0]                     cii_dout_reg;

    // Any Read Data from Capabilities to override CFG RD CPLD payload
    assign cii_override_din       = virtio_cmncfg_header_data_i | virtio_cmncfg_bar_data_i | virtio_cmncfg_baroffset_data_i | virtio_cmncfg_length_data_i |
                                    virtio_notifications_header_data_i | virtio_notifications_bar_data_i | virtio_notifications_baroffset_data_i | 
                                    virtio_notifications_length_data_i | virtio_notifications_notifyoff_data_i |
                                    virtio_isrstatus_header_data_i | virtio_isrstatus_bar_data_i | virtio_isrstatus_baroffset_data_i | virtio_isrstatus_length_data_i |
                                    virtio_devspecific_header_data_i | virtio_devspecific_bar_data_i | virtio_devspecific_baroffset_data_i | virtio_devspecific_length_data_i |
                                    virtio_pcicfgaccess_header_data_i | virtio_pcicfgaccess_bar_data_i | virtio_pcicfgaccess_baroffset_data_i | 
                                    virtio_pcicfgaccess_length_data_i | virtio_pcicfgaccess_cfgdata_data_i;                             

    always @ (posedge clk) begin
        if (~rstn) begin 
            virtio_cfgwr_req_ack_reg        <= 1'b0;
            virtio_cfgrd_req_ack_reg        <= 1'b0;
            noregbank_selected_reg          <= 1'b0;
            targeted_cap_not_present_reg    <= 1'b0;
        end    
        else begin
            virtio_cfgwr_req_ack_reg        <= virtio_cfgwr_req_ack_i;
            virtio_cfgrd_req_ack_reg        <= virtio_cfgrd_req_ack_i;
            noregbank_selected_reg          <= noregbank_selected_i;
            targeted_cap_not_present_reg    <= targeted_cap_not_present_i;
        end
    end  
     
    always @ (posedge clk) begin
        // Initialize override_en and halt registers
        if (~rstn) begin
            cii_override_din_reg <= 32'b0;
            cii_override_en_reg  <= 1'b0; 
            cii_halt_reg         <= 1'b1;   // tie to 1 as long as CII is enabled
            cii_ctrl_state       <= IDLE;
        end
        // State Machine
        else begin
            case (cii_ctrl_state)
                IDLE :
                begin
                    cii_override_din_reg <= 32'b0;
                    cii_override_en_reg  <= 1'b0; 
                    cii_halt_reg         <= 1'b1;  // this will hold req if there comes a req
                    // Detect rising edge
                    if ((noregbank_selected_i & !noregbank_selected_reg) |
                        (targeted_cap_not_present_i & !targeted_cap_not_present_reg)) begin
                        cii_ctrl_state       <= NOHIT;
                    end
                    // If there's a virtio reg hit, and it's wr req to a RO reg, same as NOHIT
                    // Will do nothing and release the halt. On HIP side, SC will be sent back.
           //         else if (!noregbank_selected_i & (virtio_cfgwr_req_ack_i & !virtio_cfgwr_req_ack_reg)) begin  // rising edge
           //             cii_ctrl_state       <= NOHIT;
           //         end
                    // If there's a legal virtio req hit, then detect ack rising edge
                    else if (!noregbank_selected_i &  ((virtio_cfgwr_req_ack_i & !virtio_cfgwr_req_ack_reg) | 
                                                       (virtio_cfgrd_req_ack_i & !virtio_cfgrd_req_ack_reg))) begin
                        cii_ctrl_state       <= CFGREQ;
                    end
                    // Wait for next req
                    else begin
                        cii_ctrl_state       <= IDLE;
                    end
                end
                CFGREQ:
                begin
                    if (virtio_cfgwr_req_ack_i) begin  
                        // De-assert halt for one cycle, but won't assert override_en
                        cii_override_din_reg <= 32'b0;
                        cii_override_en_reg  <= 1'b0; 
                        cii_halt_reg         <= 1'b0;
                    end
                    else if (virtio_cfgrd_req_ack_i) begin  
                        // De-assert halt for one cycle, and also assert override_en to override payload of cfgrd cpld
                        cii_override_din_reg <= cii_override_din;
                        cii_override_en_reg  <= 1'b1; 
                        cii_halt_reg         <= 1'b0;
                    end
                    cii_ctrl_state       <= IDLE;
                end
                NOHIT:
                begin
                    // De-assert halt for one cycle to release non-virtio reqs
                    cii_override_din_reg <= pin_cii_override_din_i;
                    cii_override_en_reg  <= pin_cii_override_en_i; 
                    cii_halt_reg         <= pin_cii_halt_i;
                    if (~pin_cii_halt_i) begin
                        cii_ctrl_state       <= IDLE;  // Wait for next req
                    end else begin
                        cii_ctrl_state       <= NOHIT;
                    end
                end
                default:
                begin
                    cii_override_din_reg <= 32'b0;
                    cii_override_en_reg  <= 1'b0; 
                    cii_halt_reg         <= 1'b1;  // tie to 1 as long as CII is enabled
                    cii_ctrl_state       <= IDLE;
                end
            endcase    
        end
    end

        
    assign cii_override_en_o            = cii_override_en_reg;
    assign cii_override_din_o           = cii_override_din_reg;
    assign cii_halt_o                   = cii_halt_reg;

    always @ (posedge clk) begin
        if (~rstn) begin
            cii_req_reg          <= 'h0; 
            cii_hdr_poisoned_reg <= 'h0;
            cii_hdr_first_be_reg <= 'h0;
            cii_func_num_reg     <= 'h0;
            cii_wr_vf_active_reg <= 'h0;
            cii_vf_num_reg       <= 'h0;
            cii_wr_reg           <= 'h0;
            cii_addr_reg         <= 'h0;
            cii_dout_reg         <= 'h0;
        end else begin
            cii_req_reg          <= (~cii_req_i)? 1'b0 : (cii_ctrl_state==NOHIT)? 1'b1 : cii_req_reg;
            cii_hdr_poisoned_reg <= cii_hdr_poisoned_i;
            cii_hdr_first_be_reg <= cii_hdr_first_be_i;
            cii_func_num_reg     <= cii_func_num_i;
            cii_wr_vf_active_reg <= cii_wr_vf_active_i;
            cii_vf_num_reg       <= cii_vf_num_i;
            cii_wr_reg           <= cii_wr_i;
            cii_addr_reg         <= cii_addr_i;
            cii_dout_reg         <= cii_dout_i;
        end
    end

    assign pin_cii_req_o          = cii_req_reg;
    assign pin_cii_hdr_poisoned_o = cii_hdr_poisoned_reg;
    assign pin_cii_hdr_first_be_o = cii_hdr_first_be_reg;
    assign pin_cii_func_num_o     = cii_func_num_reg;
    assign pin_cii_wr_vf_active_o = cii_wr_vf_active_reg;
    assign pin_cii_vf_num_o       = cii_vf_num_reg;
    assign pin_cii_wr_o           = cii_wr_reg;
    assign pin_cii_addr_o         = cii_addr_reg;
    assign pin_cii_dout_o         = cii_dout_reg;

endmodule
