// (C) 2001-2020 Intel Corporation. All rights reserved.
// Your use of Intel Corporation's design tools, logic functions and other 
// software and tools, and its AMPP partner logic functions, and any output 
// files from any of the foregoing (including device programming or simulation 
// files), and any associated documentation or information are expressly subject 
// to the terms and conditions of the Intel Program License Subscription 
// Agreement, Intel FPGA IP License Agreement, or other applicable 
// license agreement, including, without limitation, that your use is for the 
// sole purpose of programming logic devices manufactured by Intel and sold by 
// Intel or its authorized distributors.  Please refer to the applicable 
// agreement for further details.


//This module is created to facilitated pulse-width check (To be added later) for synchronizer input.
//To use this as data synchronizer with data reset value=0, simply assigned parameter RESET_VALUE=0. (or simply leave it out as the default is 0 for RESET_VALUE)
//To use this as data synchronizer with data reset value=1, simply assigned parameter RESET_VALUE=1.
//To use this as active low reset synchronizer. Assign input datain to 1(constant), and parameter RESET_VALUE=0. (or simply leave it out as the default is 0 for RESET_VALUE)
//To use this as active high reset synchronizer. Assign input datain to 0(constant), and parameter RESET_VALUE=1. 

//To use this as synchronizer with registered input. You can ignore the source_clk and source_clk_rstn 
//ports connection. Set the reset value as mentioned above as data synchronizer and reset synchronizer. 
//To use this as syncrhonizer with combi input. Set the reset value as well as the COMBI_INPUT=1
//To use this as pulse extender and synchronizer. Set the PULSE_WIDTH to be greater than 1
`timescale 1ps/1ps
module sl4_synchronizer #(
   parameter PULSE_WIDTH = 1, //default is 1 which means no extension
   parameter COMBI_INPUT = 0, //default is 0: the input signal is registered signal. 
                              //1: The datain is combi signal and will be registered in this module before sync
   parameter RESET_VALUE = 0,
   parameter SYNC_DEPTH  = 3
)(
   input source_clk,
   input source_clk_rstn,
   input dest_clk,
   input dest_clk_rstn,
   input datain,
   output wire dataout
);

   wire ext, extended;
   wire extended_tmp;
   wire dataout_tmp;
   reg  pass;


   reg extended_d;
   wire extended_rise, extended_fall;
   
   //generate pulse extender here
    generate if (PULSE_WIDTH > 1 && PULSE_WIDTH == 2)
   begin
      reg [PULSE_WIDTH-2:0] ext_reg;
      reg                   ext_hit;
      always @ (posedge source_clk)
      begin
         if (!source_clk_rstn) begin
            ext_reg <= {(PULSE_WIDTH-1){1'b0}};
            ext_hit <= 1'b0;
         end else begin
            ext_reg <= datain;
            ext_hit <= ext_reg[PULSE_WIDTH-2];
         end
	  end
   
      assign ext = ext_hit ? datain : datain || (|ext_reg);
   end
   else if (PULSE_WIDTH > 1 && PULSE_WIDTH > 2)
   begin
      reg [PULSE_WIDTH-2:0] ext_reg;
      reg                   ext_hit;
      always @ (posedge source_clk)
      begin
         if (!source_clk_rstn) begin
            ext_reg <= {(PULSE_WIDTH-1){1'b0}};
            ext_hit <= 1'b0;
         end else begin
            ext_reg <= {ext_reg[PULSE_WIDTH-3:0], datain};
            ext_hit <= ext_reg[PULSE_WIDTH-2];
         end
	  end
   
      assign ext = ext_hit ? datain : datain || (|ext_reg);
   end
   else
   begin
      assign ext = datain;
   end
   endgenerate

   //reg the extended signal that has combi before sync to dest clock
   generate if (PULSE_WIDTH > 1 || COMBI_INPUT)
   begin
      reg  extended_reg;
      always @ (posedge source_clk)
      begin
         if (!source_clk_rstn)
            extended_reg <= 1'b0;
         else
            extended_reg <= ext;
      end
      assign extended = extended_reg;
   end
   else
   begin
      assign extended = ext;
   end
   endgenerate

	//synthesis translate_off
   time period1, period2, period_inp;
   time dest_clk_smp1, dest_clk_smp2, dest_clk_prd;
    initial begin
        if ((RESET_VALUE != 0) && (RESET_VALUE != 1)) begin
            $error("%m: Error: synchronizer reset value has to be 0 or 1: %0d is not valid.", RESET_VALUE);
        end
        
        // initialize time values for assertion to 0
        period2       = 0;
        period1       = 0;
        period_inp    = 0; 
        dest_clk_smp1 = 0;
        dest_clk_smp2 = 0;
        dest_clk_prd  = 0;
        pass   = 1;

    end
	  
	  assign extended_rise = extended && !extended_d;
	  assign extended_fall = !extended && extended_d;
	//Pulse width check to be added
    // flag input extended pulse that are < 2*dest_clk
       always @ (*)
       begin
          if (dest_clk_rstn)
          begin
             if (extended_rise)
                period1 = $time;
             else if (extended_fall)
                period2 = $time;
             period_inp = (period2 > period1)? (period2 - period1): (period1 - period2);
          end
       end
       always @ (posedge dest_clk) 
          begin
             dest_clk_smp2 = $time;
             dest_clk_prd  = dest_clk_smp2 - dest_clk_smp1 -1; // minus 1ps due to simulator rounding up issue
             dest_clk_smp1 = dest_clk_smp2;
             extended_d    = extended;
          end
       always @ (period_inp)
       begin
          // if not bypassed from simpopts
          if(!$test$plusargs("SL4_SYNCH_NOCHECK"))
          begin
             //$display("Enabled Synchronizer pulse checking");   // for debug
             if (dest_clk_rstn && ((PULSE_WIDTH < 2) || (source_clk_rstn && PULSE_WIDTH>1))) begin 
                if ( period_inp < 2*dest_clk_prd) begin
                   pass = 0;
                   //$error("%m: Error: Synchronizer input(extended) pulse must be more or equal to 2 clock cycle of destination clock");
                end
                else begin
                   pass = 1;
                end
             end
          end
          else begin
             pass = 1;
          end
       end
	
	//synthesis translate_on  

	            altera_std_synchronizer_nocut #(.depth(SYNC_DEPTH))
                                    u  (
                                        .clk(dest_clk), 
                                        .reset_n(dest_clk_rstn), 
                                        .din(extended_tmp), 
                                        .dout(dataout_tmp)
                                        );
	
	assign dataout = (RESET_VALUE == 0) ? dataout_tmp : ~dataout_tmp;
	assign extended_tmp = (RESET_VALUE == 0) ? extended : ~extended;

endmodule // sl4_synchronizer
