// (C) 2001-2020 Intel Corporation. All rights reserved.
// Your use of Intel Corporation's design tools, logic functions and other 
// software and tools, and its AMPP partner logic functions, and any output 
// files from any of the foregoing (including device programming or simulation 
// files), and any associated documentation or information are expressly subject 
// to the terms and conditions of the Intel Program License Subscription 
// Agreement, Intel FPGA IP License Agreement, or other applicable 
// license agreement, including, without limitation, that your use is for the 
// sole purpose of programming logic devices manufactured by Intel and sold by 
// Intel or its authorized distributors.  Please refer to the applicable 
// agreement for further details.


`timescale 1ns/1ns


module seriallite4_source_adaptation_ecc #
(
   parameter       DEVICE_FAMILY          = "Stratix 10",
   parameter       ADAPT_PFULL_THRESHOLD  = 15,
   parameter       LANES                  = 4
)
(
   // Clocks and reset
   input                                       tx_user_clock,
   input                                       tx_user_clock_reset,
   input                                       txcore_clock,
   input                                       txcore_clock_reset,

   // Framing Interface
   input       [(LANES*64+32)-1:0]             fi_data,
   input                                       fi_valid,

   // PHY Interface
   input                                       tx_fifo_pfull,
   output reg   [(LANES*64+32)-1:0]             tx_parallel_data /* synthesis preserve_syn_only */,
   output wire  [LANES-1:0]                     tx_data_valid,

   // Error indication
   output      [LANES-1:0]                     tx_ecc_cor,
   output      [LANES-1:0]                     tx_ecc_fatal,
   output wire [LANES-1:0]                     adaptation_fifo_pfull,
   output      [LANES-1:0]                     adaptation_fifo_full
);

   localparam TARGET_CHIP = (DEVICE_FAMILY == "Stratix 10") ? 6 : 7;
   localparam PEMPTY_PART = 24;
   localparam PFULL_PART  = 6;
   localparam WAIT        = 3;
   localparam STAGES      = WAIT;
   localparam SPTRW       = 3;
   localparam PIPE_WIDTH   = (LANES*64+32)*7;
   localparam INIT        = 2'b00;
   localparam VALID       = 2'b01;
   localparam INVALID     = 2'b10;   

   reg                                         valid;
   reg                                         valid_dout;
   reg                                         valid_dout_d;
   reg                                         valid_cnt;
   reg                                         tx_data_valid_d;
   reg                                         tx_fifo_pfull_d;
   reg                                         dout_valid_d;
   wire                                        adaptation_fifo_empty;
   wire                                        tx_fifo_read;
   wire                                        adaptation_fifo_write;
   wire                                        adaptation_fifo_read;
   wire        [WAIT:0]                        read_pipe_nxt;
   reg         [WAIT-1:0]                      read_pipe;
   wire                                        dout_valid_nxt;
   reg         [3:0]                           dout_valid /* synthesis preserve_syn_only */;
   wire        [1:0]                           stage_cnt_nxt;
   reg         [7:0]                           stage_cnt /* synthesis preserve_syn_only */;

   reg         [((LANES*64+32)*(4+1))-1:0]     dout_pipe_d;
   reg         [((LANES*64+32)*4)-1:0]         dout_pipe;
   wire        [(LANES*64+32)-1:0]             adaptation_fifo_dout;
   wire        [(LANES*64+32)-1:0]             scfifo_dout;
   reg         [(LANES*64+32)-1:0]             adaptation_fifo_dout_d;
   wire                           tx_ecc_cor_combined;
   wire                           tx_ecc_fatal_combined;
   wire                           adaptation_fifo_pfull_combined;
   wire                           adaptation_fifo_full_combined;
   wire                           enable_add_cnt;
   wire                           enable_minus_cnt;
   wire                           scfifo_pfull;
   wire                           scfifo_wrreq;
   wire                           scfifo_rdreq;
   wire                           scfifo_empty;
   wire        [2:0]              rdusedw;
   reg                            scfifo_fifo_valid;
   reg                            pipe_valid_d;
   reg                            enable_minus_cnt_d;
   reg                            pipe_valid;
   reg                            valid_stage_cnt;
   reg         [1:0]              state;
   reg         [1:0]              next_state;
   integer lane;
   
   assign tx_ecc_cor                = adaptation_fifo_empty ? {LANES{1'b0}} : {LANES{tx_ecc_cor_combined}};
   assign tx_ecc_fatal              = adaptation_fifo_empty ? {LANES{1'b0}} : {LANES{tx_ecc_fatal_combined}};
   assign adaptation_fifo_pfull     = {LANES{adaptation_fifo_pfull_combined}};
   assign adaptation_fifo_full      = {LANES{adaptation_fifo_full_combined}};
   assign tx_fifo_read              = tx_fifo_pfull ? 1'b1 : 1'b0; //tx_avs_ready from tx_mac, when core ready , read
   assign adaptation_fifo_write     = fi_valid;
   assign adaptation_fifo_read      = ~scfifo_pfull & ~adaptation_fifo_empty;
   assign read_pipe_nxt             = {read_pipe, adaptation_fifo_read};
   assign dout_valid_nxt            = read_pipe_nxt[WAIT-1];
   assign enable_add_cnt            = dout_valid[0] & ~tx_fifo_pfull;
   assign enable_minus_cnt          = ~dout_valid[0] & tx_fifo_pfull; 													   
   assign stage_cnt_nxt             = enable_add_cnt && pipe_valid ? stage_cnt[1:0] + 2'b1 : 
                                      enable_minus_cnt ? valid_stage_cnt ? stage_cnt[1:0] - 2'b1 : stage_cnt[1:0] : stage_cnt[1:0];
   assign scfifo_pfull              = (rdusedw <= 3) ? 1'b0 : 1'b1;
   assign scfifo_wrreq              = dout_valid[0]; 
   assign scfifo_rdreq              = tx_fifo_pfull & ~scfifo_empty;        
 
   always @ (posedge txcore_clock) begin
       if (txcore_clock_reset) begin
	   state            <= INIT;
           pipe_valid_d     <= 1'b0;
           dout_valid_d     <= 1'b0;
           tx_data_valid_d  <= 1'b0;
           valid_stage_cnt  <= 1'b0; 
       end else begin
	   state            <= next_state;
           pipe_valid_d     <= pipe_valid;
           dout_valid_d     <= dout_valid[1];
           tx_data_valid_d  <= tx_data_valid[0];
           valid_stage_cnt  <= |stage_cnt_nxt;
       end
   end		 


   always @ (posedge txcore_clock) begin
      for (lane=0; lane<=STAGES;lane=lane+1) begin
         if (txcore_clock_reset) begin
            stage_cnt[lane*2 +: 2]   <= 2'b0;
            dout_valid[lane]         <= 1'b0;
         end else begin
            stage_cnt[lane*2 +: 2]   <= stage_cnt_nxt;
            dout_valid[lane]         <= read_pipe_nxt[2];
         end			
      end
   end	  
		   
   always @ (*) begin
       case (state)
           INIT : begin
               if (dout_valid[2] && ~dout_valid_d) begin
                   next_state      <= VALID;
                   pipe_valid      <= 1'b0;	
               end else begin
                   next_state      <= INIT;
                   pipe_valid      <= 1'b0;	   
               end									
           end
           VALID : begin
                if (valid_stage_cnt || ~tx_fifo_pfull) begin
                    next_state      <= VALID;
                    pipe_valid      <= 1'b1;				
                end else begin
                    next_state  <= INVALID;
                    pipe_valid  <= 1'b0;
                end
           end
           INVALID : begin
               if (~tx_fifo_pfull && ((dout_valid[2] && dout_valid_d) || (valid_dout_d))) begin
                   next_state  <= VALID;		
                   pipe_valid  <= 1'b1;
               end else begin
                   next_state  <= INVALID;
                   pipe_valid  <= pipe_valid_d;
               end
           end      
	    default : begin next_state <= INIT;  end           
            endcase
	end
	
    
   always @ (posedge txcore_clock) begin
      if (txcore_clock_reset) begin
         read_pipe                 <= {WAIT{1'b0}};
         valid                     <= 1'b0;
         valid_dout_d              <= 1'b0;
         dout_pipe_d               <= {PIPE_WIDTH{1'b0}};
         //valid_cnt                 <= 1'b0;
         enable_minus_cnt_d        <= 1'b0;		
         scfifo_fifo_valid         <= 1'b0;		 
      end
      else begin
         read_pipe                 <= read_pipe_nxt[WAIT-1:0];
         valid                     <= 1'b1;
         valid_dout_d              <= (valid_dout_d || valid_cnt) ? ~tx_fifo_pfull : 1'b0;
         dout_pipe_d               <= {dout_pipe_d[((LANES*64+32)*0) +: (LANES*64+32)], dout_pipe};
         //valid_cnt                 <= (dout_valid[2]) ? 1'b1 : |stage_cnt[1:0] ? 1'b1 : ((valid_dout_d || valid_cnt) && ~tx_fifo_pfull);
         enable_minus_cnt_d        <= enable_minus_cnt;
         scfifo_fifo_valid         <= scfifo_rdreq;		
      end
   end

   //*********************************************************************
   //
   // Per-Lane Logic
   //
   //*********************************************************************
   wire [5-1:0] wused;

         //
         // Lane rate adaptation FIFO instantiation
         //
            // ECC Enabled FIFOs
            dcfifo_s5m20k #
            (
               .WIDTH        (64*LANES+32),
               .DISABLE_WUSED(1'b0),
               .DISABLE_RUSED(1'b1),
               .PREVENT_UNDERFLOW(1'b0),
               .TARGET_CHIP  (TARGET_CHIP),
               .ECC_PIPE     ("TRUE")
            )
            dcfifo_ecc
            (
               .aclr       (tx_user_clock_reset),
               .sclr_err   (1'b0),
               .wclk       (tx_user_clock),
               .wdata      (fi_data),
               .wreq       (adaptation_fifo_write),
               .wfull      (adaptation_fifo_full_combined),
               .wused      (wused),
               .rclk       (txcore_clock),
               .rdata      (adaptation_fifo_dout),
               .rreq       (adaptation_fifo_read),
               .rempty     (adaptation_fifo_empty),
               .ecc_status ({tx_ecc_fatal_combined, tx_ecc_cor_combined})
            );
         
            assign adaptation_fifo_pfull_combined = (wused > ADAPT_PFULL_THRESHOLD[4:0]);
   					
			scfifo  #(
                .add_ram_output_register ("ON"),
                .enable_ecc             ("FALSE"),
                .intended_device_family (DEVICE_FAMILY),
                .lpm_numwords           (8),
                .lpm_showahead          ("OFF"),
                .lpm_hint               ("RAM_BLOCK_TYPE=MLAB"),
                .lpm_type               ("scfifo"),
                .lpm_width              (LANES*64+32),
                .lpm_widthu             (3),
                .overflow_checking      ("ON"),
                .underflow_checking     ("ON"),
                .use_eab                ("ON")
	        ) rx_scfifo (
                .sclr                   ( txcore_clock_reset                                           ), // input
                .aclr                   (1'b0                                                          ),
                .clock                  ( txcore_clock                                                 ), // input
                .wrreq                  ( scfifo_wrreq                                                 ), // input
                .data                   ( adaptation_fifo_dout                                         ),
                .rdreq                  ( scfifo_rdreq                                                 ), // input
                .q                      ( scfifo_dout                                                  ), 
                .empty                  ( scfifo_empty                                                 ),
                .usedw                  ( rdusedw                                                      )  // output [2:0]
	       );

           ready_skid #(
	           .WIDTH (LANES*64+32)
           ) ready_skid (
               .clk                     (txcore_clock       ),
               .sclr                    (txcore_clock_reset ),
               .valid_i                 (scfifo_fifo_valid  ),
               .dat_i                   (scfifo_dout        ),
               .ready_o                 (tx_fifo_pfull      ),
	
               .valid_o                 (valid_cnt          ),
               .dat_o                   (tx_parallel_data   )
           );

   assign tx_data_valid             = (tx_fifo_pfull && valid_cnt) ? {(LANES){1'b1}} : {(LANES){1'b0}};    
                                            											
endmodule
